//#include "ssl_locl.h"

#ifndef STATE_MACHINE_H_
#define STATE_MACHINE_H_

/*@
 axiomatic StateMachine {

//===================================================================
// 				Predicates on who sent the message
//===================================================================

 predicate sentByClient(STATE_MON *state, int direction) =
 	 (state->role == SERVER && direction == RECEIVE)||
 	 (state->role == CLIENT && direction == SEND);

 predicate sentByServer(STATE_MON *state, int direction) =
 	 (state->role == CLIENT && direction == RECEIVE)||
 	 (state->role == SERVER && direction == SEND);

//===================================================================
// 				Predicates on the nature of the message
//===================================================================

// predicate isMessageLength(unsigned char *msg, unsigned int len) =
// 	 len >= 4 && (len - 4) == (65536 * msg[1] + 256 * msg[2] + msg[3]);
predicate isMessageLength(unsigned char *msg, unsigned int len) =
 	 len >= 4 && (len-4) == (65536 * msg[1] + 256 * msg[2] + msg[3]);

 predicate isClientHello(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == C_Hello &&
 	 isMessageLength(msg, len);
 predicate isServerHello(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == S_Hello &&
 	 isMessageLength(msg, len);
 predicate isServerCertificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == handshake && msg[0] == certificate && sentByServer(state, direction)  &&
 	 isMessageLength(msg, len);
 predicate isServerKeyExchange(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == S_KE &&
 	 isMessageLength(msg, len);
 predicate isServerCertificateRequest(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == S_CReq &&
 	 isMessageLength(msg, len);
 predicate isServerHelloDone(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == S_HD &&
 	 isMessageLength(msg, len);
 predicate isClientCertificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == handshake && msg[0] == certificate && sentByClient(state, direction)  &&
 	 isMessageLength(msg, len);
 predicate isClientKeyExchange(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == C_KE &&
 	 isMessageLength(msg, len);
 predicate isClientCertificateVerify(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == C_CV &&
 	 isMessageLength(msg, len);
 predicate isServerChangeCipherSpecs(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == change_cipher_spec && sentByServer(state, direction) &&
 	 len == 1 && msg[0] == 0x01;
 predicate isServerFin(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == handshake && msg[0] == finished && sentByServer(state, direction)  &&
 	 isMessageLength(msg, len);
 predicate isClientChangeCipherSpecs(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == change_cipher_spec && sentByClient(state, direction) &&
 	 len == 1 && msg[0] == 0x01;
 predicate isClientFin(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == handshake && msg[0] == finished && sentByClient(state, direction) &&
 	 isMessageLength(msg, len);
 predicate isServerNewSessionTicket(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == ServerNewSessionTicket &&
 	 isMessageLength(msg, len);

 predicate isUnknownMessage(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 !isClientHello(msg, len, contentType) 									&&
 	 !isServerHello(msg, len, contentType) 									&&
	 !isServerCertificate(state, msg, len, direction, contentType) 			&&
	 !isServerKeyExchange(msg, len, contentType) 							&&
	 !isServerCertificateRequest(msg, len, contentType) 					&&
	 !isServerHelloDone(msg, len, contentType) 								&&
	 !isClientCertificate(state, msg, len, direction, contentType) 			&&
	 !isClientKeyExchange(msg, len, contentType) 							&&
	 !isClientCertificateVerify(msg, len, contentType) 						&&
	 !isServerChangeCipherSpecs(state, msg, len, direction, contentType) 	&&
	 !isServerFin(state, msg, len, direction, contentType) 					&&
	 !isClientChangeCipherSpecs(state, msg, len, direction, contentType) 	&&
	 !isServerNewSessionTicket(msg, len, contentType) 						&&
	 !isClientFin(state, msg, len, direction, contentType)					;

 predicate isEmptyClientCertificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction) =
 	 isClientCertificate(state, msg, len, direction, handshake) 		&&
 	 len == 4;

 predicate isNonEmptyClientCertificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction) =
 	 isClientCertificate(state, msg, len, direction, handshake)			&&
 	 len > 4;

//===================================================================
// 				Predicates on the protocol versions
//===================================================================

 predicate isSSL3{L}(unsigned char *msg, unsigned int len)reads msg+(0..len-1);
 predicate isTLS1_0{L}(unsigned char *msg, unsigned int len)reads msg+(0..len-1);
 predicate isTLS1_1{L}(unsigned char *msg, unsigned int len)reads msg+(0..len-1);
 predicate isTLS1_2{L}(unsigned char *msg, unsigned int len)reads msg+(0..len-1);

 predicate isUnknownPV{L}(unsigned char *msg, unsigned int len) =
 	 !isSSL3(msg, len) 		&&
 	 !isTLS1_0(msg, len) 	&&
 	 !isTLS1_1(msg, len) 	&&
 	 !isTLS1_2(msg, len)	;

//===================================================================
// 				Predicates on the key exchange methods
//===================================================================

 predicate isDH{L}(unsigned char *msg, unsigned int len) reads msg+(0..len-1);
 predicate isDHE{L}(unsigned char *msg, unsigned int len) reads msg+(0..len-1);
 predicate isDHEExport{L}(unsigned char *msg, unsigned int len) reads msg+(0..len-1);
 predicate isDHanon{L}(unsigned char *msg, unsigned int len) reads msg+(0..len-1);
 predicate isECDH{L}(unsigned char *msg, unsigned int len) reads msg+(0..len-1);
 predicate isECDHE{L}(unsigned char *msg, unsigned int len) reads msg+(0..len-1);
 predicate isECDHanon{L}(unsigned char *msg, unsigned int len) reads msg+(0..len-1);
 predicate isRSA{L}(unsigned char *msg, unsigned int len) reads msg+(0..len-1);
 predicate isRSAExport{L}(unsigned char *msg, unsigned int len) reads msg+(0..len-1);

 predicate isUnknownKEM{L}(unsigned char *msg, unsigned int len)  =
 	 !isDH(msg, len) 			&&
 	 !isDHE(msg, len)			&&
 	 !isDHEExport(msg, len)		&&
 	 !isDHanon(msg, len) 		&&
 	 !isECDH(msg, len) 			&&
 	 !isECDHE(msg, len) 		&&
 	 !isECDHanon(msg, len) 		&&
 	 !isRSA(msg, len) 			&&
 	 !isRSAExport(msg, len) 	;

//===================================================================
// 				Log predicates
//===================================================================

 predicate MessageAddedToLog_E{L}(STATE_MON *new_state, STATE_MON *previous_state, unsigned char *message, unsigned int len) =
 	new_state->log_length == previous_state->log_length + len &&
 	(\forall integer i; 0 <= i < previous_state->log_length ==> previous_state->log[i] == new_state->log[i]) &&
	(\forall integer i; 0 <= i < len ==> new_state->log[i+previous_state->log_length] == message[i]);

 predicate MessageAddedToLog{New,Old}(STATE_MON *state, unsigned char *msg, unsigned int len) =
 	 \at(state->log_length,Old) + len == \at(state->log_length,New) &&
	 (\forall integer i; 0 <= i < \at(state->log_length,Old) ==> \at(state->log[i],Old) == \at(state->log[i],New)) &&
	 (\forall integer i; 0 <= i < len ==> \at(state->log[i+\at(state->log_length,Old)],New) == \at(msg[i],Old)) ;


//===================================================================
// 				State monitor update predicates
//===================================================================

 predicate AgreeOnVersion{L}(STATE_MON *s, unsigned char *msg, unsigned int len) =
 	 (\at(s->version,L) == SSL3 && isSSL3(msg, len)) 			||
 	 (\at(s->version,L) == TLS1 && isTLS1_0(msg, len)) 			||
 	 (\at(s->version,L) == TLS1_1 && isTLS1_1(msg, len)) 		||
 	 (\at(s->version,L) == TLS1_2 && isTLS1_2(msg, len)) 		;

 predicate AgreeOnKEM{L}(STATE_MON *s, unsigned char *msg, unsigned int len) =
 	 (\at(s->kem,L) == DiffieHellman && isDH(msg, len)) 		||
 	 (\at(s->kem,L) == DHE && isDHE(msg, len))					||
 	 (\at(s->kem,L) == DHEExport && isDHEExport(msg, len))		||
 	 (\at(s->kem,L) == DHanon && isDHanon(msg, len)) 			||
 	 (\at(s->kem,L) == ECDH && isECDH(msg, len)) 				||
 	 (\at(s->kem,L) == ECDHE && isECDHE(msg, len)) 				||
 	 (\at(s->kem,L) == ECDHanon && isECDHanon(msg, len)) 		||
 	 (\at(s->kem,L) == CS_RSA && isRSA(msg, len)) 				||
 	 (\at(s->kem,L) == RSAExport && isRSAExport(msg, len)) 		;

 predicate CsRequiresServerCertificate{L}(STATE_MON *s) =
 	 (\at(s->kem,L) == DiffieHellman) 		||
 	 (\at(s->kem,L) == DHE) 				||
 	 (\at(s->kem,L) == DHEExport)			||
 	 (\at(s->kem,L) == ECDH) 				||
 	 (\at(s->kem,L) == ECDHE ) 				||
 	 (\at(s->kem,L) == RSAExport ) 			||
 	 (\at(s->kem,L) == CS_RSA) 				;

 predicate CsRequiresServerKeyExchange{L}(STATE_MON *s) =
 	 (\at(s->kem,L) == DHE) 				||
 	 (\at(s->kem,L) == DHEExport) 			||
 	 (\at(s->kem,L) == RSAExport) 			||
 	 (\at(s->kem,L) == ECDHE ) 				||
 	 (\at(s->kem,L) == DHanon ) 			||
 	 (\at(s->kem,L) == ECDHanon) 			;

 predicate HaveSameStateValuesAndRole{L1,L2}(STATE_MON *s) =
	\at(s->version,L1) 			== \at(s->version,L2) 				&&
	\at(s->role,L1) 			== \at(s->role,L2) 					&&
	\at(s->kem,L1) 				== \at(s->kem,L2) 					&&
	\at(s->client_auth,L1) 		== \at(s->client_auth,L2) 			&&
	\at(s->resumption,L1) 		== \at(s->resumption,L2) 			&&
	\at(s->renegotiation,L1) 	== \at(s->renegotiation,L2) 		&&
	\at(s->ntick, L1) 			== \at(s->ntick, L2)				;

 predicate HaveSameStateValuesAndRole_E{L}(STATE_MON *s1, STATE_MON *s2) =
	\at(s1->version,L) 			== \at(s2->version,L) 				&&
	\at(s1->role,L) 			== \at(s2->role,L) 					&&
	\at(s1->kem,L) 				== \at(s2->kem,L) 					&&
	\at(s1->client_auth,L) 		== \at(s2->client_auth,L) 			&&
	\at(s1->resumption,L) 		== \at(s2->resumption,L) 			&&
	\at(s1->renegotiation,L) 	== \at(s2->renegotiation,L) 		&&
	\at(s1->ntick, L) 			== \at(s2->ntick, L)				;

 predicate HaveSameStateValuesButVersionAndCs{L1,L2}(STATE_MON *s) =
	\at(s->role,L1) 			== \at(s->role,L2) 					&&
	\at(s->client_auth,L1) 		== \at(s->client_auth,L2) 			&&
	\at(s->resumption,L1)		== \at(s->resumption,L2) 			&&
	\at(s->renegotiation,L1) 	== \at(s->renegotiation,L2) 		&&
	\at(s->ntick, L1) 			== \at(s->ntick, L2)				;

 predicate HaveSameStateValuesButVersionAndCs_E{L}(STATE_MON *s1, STATE_MON *s2) =
	\at(s1->role,L) 			== \at(s2->role,L) 					&&
	\at(s1->client_auth,L) 		== \at(s2->client_auth,L) 			&&
	\at(s1->resumption,L) 		== \at(s2->resumption,L) 			&&
	\at(s1->renegotiation,L) 	== \at(s2->renegotiation,L) 		&&
	\at(s1->ntick, L) 			== \at(s2->ntick, L)				;


  predicate HaveSameStateValuesButRole{L1,L2}(STATE_MON *s) =
	\at(s->version,L1) 			== \at(s->version,L2) 				&&
	\at(s->kem,L1) 				== \at(s->kem,L2) 					&&
	\at(s->client_auth,L1) 		== \at(s->client_auth,L2) 			&&
	\at(s->resumption,L1) 		== \at(s->resumption,L2) 			&&
	\at(s->renegotiation,L1) 	== \at(s->renegotiation,L2) 		&&
	\at(s->ntick, L1) 			== \at(s->ntick, L2)				;

 predicate HaveSameStateValuesButRole_E{L}(STATE_MON *s1, STATE_MON *s2) =
	\at(s1->version,L) 			== \at(s2->version,L) 				&&
	\at(s1->kem,L) 				== \at(s2->kem,L) 					&&
	\at(s1->client_auth,L) 		== \at(s2->client_auth,L) 			&&
	\at(s1->resumption,L) 		== \at(s2->resumption,L) 			&&
	\at(s1->renegotiation,L) 	== \at(s2->renegotiation,L)			&&
	\at(s1->ntick, L) 			== \at(s2->ntick, L)				;

 predicate HaveSameStateValuesButClientAuth{L1,L2}(STATE_MON *s) =
	\at(s->version,L1) 			== \at(s->version,L2) 				&&
	\at(s->role,L1) 			== \at(s->role,L2) 					&&
	\at(s->kem,L1) 				== \at(s->kem,L2) 					&&
	\at(s->resumption,L1) 		== \at(s->resumption,L2) 			&&
	\at(s->renegotiation,L1) 	== \at(s->renegotiation,L2)			&&
	\at(s->ntick, L1) 			== \at(s->ntick, L2)				;

 predicate HaveSameStateValuesButClientAuth_E{L}(STATE_MON *s1, STATE_MON *s2) =
	\at(s1->version,L) 			== \at(s2->version,L) 				&&
	\at(s1->role,L) 			== \at(s2->role,L) 					&&
	\at(s1->kem,L) 				== \at(s2->kem,L) 					&&
	\at(s1->resumption,L) 		== \at(s2->resumption,L) 			&&
	\at(s1->renegotiation,L) 	== \at(s2->renegotiation,L) 		&&
	\at(s1->ntick, L) 			== \at(s2->ntick, L)				;

 predicate HaveSameStateValuesButRenegociation{L1,L2}(STATE_MON *s) =
	\at(s->version,L1) 			== \at(s->version,L2) 				&&
	\at(s->role,L1) 			== \at(s->role,L2) 					&&
	\at(s->kem,L1) 				== \at(s->kem,L2) 					&&
	\at(s->client_auth,L1) 		== \at(s->client_auth,L2) 			&&
	\at(s->resumption,L1)	 	== \at(s->resumption,L2)			&&
	\at(s->ntick, L1) 			== \at(s->ntick, L2)				;

 predicate HaveSameStateValuesButResumption{L1,L2}(STATE_MON *s) =
	\at(s->version,L1) 			== \at(s->version,L2) 				&&
	\at(s->role,L1) 			== \at(s->role,L2) 					&&
	\at(s->kem,L1) 				== \at(s->kem,L2) 					&&
	\at(s->client_auth,L1) 		== \at(s->client_auth,L2) 			&&
	\at(s->renegotiation,L1) 	== \at(s->renegotiation,L2)			&&
	\at(s->ntick, L1) 			== \at(s->ntick, L2)				;

 predicate HaveSameStateValuesButResumption_E{L}(STATE_MON *s1, STATE_MON *s2) =
	\at(s1->version,L) 			== \at(s2->version,L) 			 	&&
	\at(s1->role,L) 			== \at(s2->role,L) 					&&
	\at(s1->kem,L) 				== \at(s2->kem,L) 					&&
	\at(s1->client_auth,L) 		== \at(s2->client_auth,L) 			&&
	\at(s1->renegotiation,L) 	== \at(s2->renegotiation,L)			&&
	\at(s1->ntick, L)			== \at(s2->ntick, L)				;

 predicate HaveSameStateValuesButResumptionAndTicket{L1,L2}(STATE_MON *s) =
	\at(s->version,L1) 			== \at(s->version,L2) 				&&
	\at(s->role,L1) 			== \at(s->role,L2) 					&&
	\at(s->kem,L1) 				== \at(s->kem,L2) 					&&
	\at(s->client_auth,L1) 		== \at(s->client_auth,L2) 			&&
	\at(s->renegotiation,L1) 	== \at(s->renegotiation,L2) 		;

 predicate HaveSameStateValuesButResumptionAndTicket_E{L}(STATE_MON *s1, STATE_MON *s2) =
	\at(s1->version,L) 			== \at(s2->version,L) 				&&
	\at(s1->role,L) 			== \at(s2->role,L) 					&&
	\at(s1->kem,L) 				== \at(s2->kem,L) 					&&
	\at(s1->client_auth,L) 		== \at(s2->client_auth,L) 			&&
	\at(s1->renegotiation,L) 	== \at(s2->renegotiation,L) 		;

 predicate HaveSameStateValuesButNewTicket{L1,L2}(STATE_MON *s) =
	\at(s->version,L1) 			== \at(s->version,L2) 				&&
	\at(s->role,L1) 			== \at(s->role,L2) 					&&
	\at(s->kem,L1) 				== \at(s->kem,L2) 					&&
	\at(s->client_auth,L1) 		== \at(s->client_auth,L2) 			&&
	\at(s->resumption,L1) 		== \at(s->resumption,L2) 			&&
	\at(s->renegotiation,L1) 	== \at(s->renegotiation,L2)			;

predicate DeepStateEquality(STATE_MON *s1, STATE_MON *s2) =
	s1->version 			== s2->version 							&&
	s1->role 				== s2->role 							&&
	s1->kem 					== s2->kem 								&&
	s1->last_message		== s2->last_message						&&
	s1->log_length 			== s2->log_length 						&&
	s1->client_auth 		== s2->client_auth 						&&
	s1->resumption 			== s2->resumption 						&&
	s1->renegotiation 		== s2->renegotiation 					&&
	s1->ntick 				== s2->ntick							&&
	s1->state_mon_initialized == s2->state_mon_initialized			&&
	\forall integer i; 0 <= i < s1->log_length ==> s1->log[i] == s2->log[i] 	;

 predicate DeepStateEquality2{L,L2}(STATE_MON *s1, STATE_MON *s2) =
	\at(s1->version,L) 			== \at(s2->version,L2) 						&&
	\at(s1->role,L) 			== \at(s2->role,L2) 						&&
	\at(s1->kem,L) 				== \at(s2->kem,L2) 							&&
	\at(s1->last_message,L)		== \at(s2->last_message,L2)					&&
	\at(s1->log_length, L) 		== \at(s2->log_length, L2) 					&&
	\at(s1->client_auth,L) 		== \at(s2->client_auth,L2) 					&&
	\at(s1->resumption,L) 		== \at(s2->resumption,L2) 					&&
	\at(s1->renegotiation,L) 	== \at(s2->renegotiation,L2) 				&&
	\at(s1->ntick, L) 			== \at(s2->ntick, L2)						&&
	\at(s1->state_mon_initialized,L) == \at(s2->state_mon_initialized,L2)	&&
	\forall integer i; 0 <= i < \at(s1->log_length,L) ==> \at(s1->log[i],L) == \at(s2->log[i],L2);

//===================================================================
// 				State monitor predicates
//===================================================================

 predicate StateAfterInitialState{L}(STATE_MON *state) =
 	 \at(state->version,L) == UNDEFINED_PV 										&&
 	 \at(state->role,L) == UNDEFINED_ROLE 										&&
 	 \at(state->kem,L) == UNDEFINED_KEM 										&&
 	 \at(state->last_message,L) == UNDEFINED_TYPE 								&&
 	 \at(state->log_length,L) == 0 												&&
 	 \at(state->client_auth,L) == UNDEFINED_AUTH 								&&
 	 \at(state->resumption,L) == UNDEFINED 										&&
 	 \at(state->renegotiation,L) == UNDEFINED									&&
 	 \at(state->ntick,L) == UNDEFINED											;

 predicate StateAfterClientHello{L}(STATE_MON *state) =
 	 \exists STATE_MON *s, unsigned char *message, unsigned int len;
		  \at(state->last_message,L) == C_Hello 								&&
		  HaveSameStateValuesButRole_E{L}(state, s) 							&&
		  (\at(state->role,L) == CLIENT || \at(state->role,L) == SERVER ) 		&&
		  MessageAddedToLog_E{L}(state, s, message, len) 						&&
		  isClientHello(message, len, handshake) 								&&
		  StateAfterInitialState(s)												;

 predicate NextStateAfterClientHello{New,Old}(STATE_MON *state, unsigned char *msg, unsigned int len) =
		  \at(state->last_message,Old) == UNDEFINED_TYPE						&&
		  \at(state->last_message,New) == C_Hello 								&&
		  HaveSameStateValuesButRole{New,Old}(state) 							&&
		  MessageAddedToLog{New,Old}(state,msg, len) 							&&
		  isClientHello{New}(msg, len, handshake) 								&&
		  StateAfterInitialState{Old}(state)									;

 predicate StateAfterServerHello{L}(STATE_MON *state) =
 	 \exists STATE_MON *s, unsigned char *message, unsigned int len;
		  \at(state->last_message,L) == S_Hello 								&&
		  HaveSameStateValuesButVersionAndCs_E{L}(state, s) 					&&
		  AgreeOnVersion{L}(state, message, len) 								&&
		  AgreeOnKEM{L}(state, message, len) 									&&
		  MessageAddedToLog_E{L}(state, s, message, len)						&&
		  isServerHello(message, len, handshake) 								&&
		  StateAfterClientHello(s)												;

 predicate NextStateAfterServerHello{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len) =
		  \at(state->last_message,New) == S_Hello 								&&
		  HaveSameStateValuesButVersionAndCs{New,Old}(state) 					&&
		  AgreeOnVersion{New}(state, message, len) 								&&
		  AgreeOnKEM{New}(state, message, len) 									&&
		  MessageAddedToLog{New,Old}(state,message, len) 						&&
		  isServerHello{New}(message, len, handshake) 							&&
		  StateAfterClientHello{Old}(state)										;

 predicate StateAfterServerCertificate{L}(STATE_MON *state) =
 	 \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		  \at(state->last_message,L) == S_Cert									&&
		  \at(state->resumption,L) == NO										&&
		  HaveSameStateValuesButResumption_E{L}(\at(state,L), s) 				&&
		  MessageAddedToLog_E(\at(state,L), s, message, len)					&&
		  isServerCertificate(\at(state,L), message, len, direction, handshake) &&
		  StateAfterServerHello(s)											;

 predicate NextStateAfterServerCertificate{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len, int direction) =
		  \at(state->last_message,New) == S_Cert 								&&
		  \at(state->resumption,New) == NO										&&
		  HaveSameStateValuesButResumption{New,Old}(state) 						&&
		  MessageAddedToLog{New,Old}(state,message, len) 						&&
		  isServerCertificate{New}(state, message, len, direction, handshake)	&&
		  StateAfterServerHello{Old}(state)									;

 predicate StateAfterServerKeyExchange{L}(STATE_MON *state) =
 	 \exists STATE_MON *s, unsigned char *message, unsigned int len;
		  \at(state->last_message,L) == S_KE 									&&
		  HaveSameStateValuesButResumption_E{L}(state, s) 						&&
		  MessageAddedToLog_E{L}(state, s, message, len)						&&
		  isServerKeyExchange{L}(message, len, handshake) 						&&
		  (
		  	  ( StateAfterServerCertificate{L}(s) && (\at(state->kem,L) == DHE || \at(state->kem,L) == ECDHE || \at(state->kem,L) == DHEExport || \at(state->kem,L) == RSAExport) && \at(state->resumption,L) == NO )
		  	  ||
		  	  ( StateAfterServerHello{L}(s) && (\at(s->kem,L) == DHanon || \at(s->kem,L) == ECDHanon) && \at(state->resumption,L) == NO )
		  )	;

 predicate NextStateAfterServerKeyExchange{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len) =
		  \at(state->last_message,New) == S_KE 									&&
		  HaveSameStateValuesButResumption{New,Old}(state) 						&&
		  MessageAddedToLog{New,Old}(state,message, len) 						&&
		  isServerKeyExchange{New}(message, len, handshake)						&&
		  (
		  	  ( StateAfterServerCertificate{Old}(state) && (\at(state->kem,Old) == DHE || \at(state->kem,Old) == DHEExport || \at(state->kem,Old) == ECDHE || \at(state->kem,Old) == RSAExport) && \at(state->resumption,New) == \at(state->resumption,Old) )
		  	  ||
		  	  ( StateAfterServerHello{Old}(state) && (\at(state->kem,Old) == DHanon || \at(state->kem,Old) == ECDHanon) && \at(state->resumption,New) == NO )
		  )	;

 predicate StateAfterServerCertificateRequest{L}(STATE_MON *state) =
 	 \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
 	 	 \at(state->last_message,L) == S_CReq 									&&
		  HaveSameStateValuesButClientAuth_E{L}(state, s) 						&&
		  \at(state->client_auth,L) == AUTH_REQUESTED							&&
		  MessageAddedToLog_E{L}(state, s, message, len) 						&&
		  isServerCertificateRequest{L}(message, len, handshake)				&&
		  (
		  	  ( StateAfterServerCertificate{L}(s) )
		  	  ||
		  	  ( StateAfterServerKeyExchange{L}(s) && (\at(state->kem,L) == DiffieHellman ||
		  	  	  	  	  	  	  	  	  	  	  	  \at(state->kem,L) == ECDH ||
		  	  	  	  	  	  	  	  	  	  	  	  \at(state->kem,L) == DHE ||
		  	  	  	  	  	  	  	  	  	  	  	  \at(state->kem,L) == DHEExport ||
		  	  	  	  	  	  	  	  	  	  	  	  \at(state->kem,L) == ECDHE ) )
		  )	;

 predicate NextStateAfterServerCertificateRequest{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len) =
 	 	 \at(state->last_message,New) == S_CReq 								&&
		  HaveSameStateValuesButClientAuth{New,Old}(state) 						&&
		  \at(state->client_auth,New) == AUTH_REQUESTED							&&
		  MessageAddedToLog{New,Old}(state,message, len) 						&&
		  isServerCertificateRequest{New}(message, len, handshake)				&&
		  (
		  	  ( StateAfterServerCertificate{Old}(state) )
		  	  ||
		  	  ( StateAfterServerKeyExchange{Old}(state) && (\at(state->kem,Old) == DiffieHellman ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == ECDH ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == DHE ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == DHEExport ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == ECDHE ) )
		  )	;

 predicate StateAfterServerHelloDone{L}(STATE_MON *state) =
  	 \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
 	 	 \at(state->last_message,L) == S_HD 		 							&&
 	 	 HaveSameStateValuesButClientAuth_E{L}(state, s) 						&&
		 MessageAddedToLog_E{L}(state, s, message, len) 						&&
		 isServerHelloDone{L}(message, len, handshake)							&&
		 (
		 	 ( StateAfterServerCertificate(s) && (s->kem == DiffieHellman ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	s->kem == ECDH ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	s->kem == CS_RSA)
		  	  	  	  	  	  	  	  	  	  	  	   && \at(s->client_auth,L) == NO_AUTH)
		  	 ||
		  	 ( StateAfterServerKeyExchange(s) && (s->kem == DHanon ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	s->kem == ECDHanon ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	s->kem == DHE ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	s->kem == DHEExport ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	s->kem == RSAExport ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	s->kem == ECDHE)
		  	  	  	  	  	  	  	  	  	  	  	  	  	&& \at(state->client_auth,L) == NO_AUTH)
		  	 ||
		  	 ( StateAfterServerCertificateRequest(s) && ( s->kem == DiffieHellman ||
																  s->kem == ECDH ||
																  s->kem == CS_RSA ||
																  s->kem == DHE ||
																  s->kem == DHEExport ||
																  s->kem == RSAExport ||
																  s->kem == ECDHE )
																  && \at(state->client_auth,L) == s->client_auth)
		 );

 predicate NextStateAfterServerHelloDone{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len) =
 	 	 \at(state->last_message,New) == S_HD 		 							&&
 	 	 HaveSameStateValuesButClientAuth{New,Old}(state) 						&&
		 MessageAddedToLog{New,Old}(state,message, len) 						&&
		 isServerHelloDone{New}(message, len, handshake)						&&
		 (
		 	 ( StateAfterServerCertificate{Old}(state) && (\at(state->kem,Old) == DiffieHellman ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == ECDH ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == CS_RSA)
		  	  	  	  	  	  	  	  	  	  	  	   && \at(state->client_auth,New) == NO_AUTH)
		  	 ||
		  	 ( StateAfterServerKeyExchange{Old}(state) && (\at(state->kem,Old) == DHanon ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == ECDHanon ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == DHE ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == DHEExport ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == RSAExport ||
		  	  	  	  	  	  	  	  	  	  	  	  	  	\at(state->kem,Old) == ECDHE)
		  	  	  	  	  	  	  	  	  	  	  	  	  	&& \at(state->client_auth,New) == NO_AUTH)
		  	 ||
		  	 ( StateAfterServerCertificateRequest{Old}(state) && (\at(state->kem,Old) == DiffieHellman ||
																  \at(state->kem,Old) == ECDH ||
																  \at(state->kem,Old) == CS_RSA ||
																  \at(state->kem,Old) == DHE ||
																  \at(state->kem,Old) == DHEExport ||
																  \at(state->kem,Old) == RSAExport ||
																  \at(state->kem,Old) == ECDHE )
																  && \at(state->client_auth,New) == \at(state->client_auth,Old))
		 );

 predicate StateAfterClientCertificate{L}(STATE_MON *state) =
   	 \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		\at(state->last_message,L) == C_Cert									&&
		HaveSameStateValuesButClientAuth_E{L}(state, s)							&&
		MessageAddedToLog_E{L}(state, s, message, len) 							&&
		isClientCertificate{L}(state, message, len, direction, handshake)		&&
		StateAfterServerHelloDone(s)											&&
		s->client_auth == AUTH_REQUESTED										&&
		(
			(isEmptyClientCertificate(s, message, len, direction) && \at(state->client_auth,L) == AUTH_NOT_PROVIDED )
			||
			(isNonEmptyClientCertificate(s, message, len, direction) && \at(state->client_auth,L) == AUTH_PROVIDED )
		);

 predicate NextStateAfterClientCertificate{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len, int direction) =
		\at(state->last_message,New) == C_Cert									&&
		HaveSameStateValuesButClientAuth{New,Old}(state)						&&
		MessageAddedToLog{New,Old}(state,message, len) 							&&
		isClientCertificate{New}(state, message, len, direction, handshake)		&&
		StateAfterServerHelloDone{Old}(state)									&&
		\at(state->client_auth,Old) == AUTH_REQUESTED							&&
		(
			(isEmptyClientCertificate{Old}(state, message, len, direction) && \at(state->client_auth,New) == AUTH_NOT_PROVIDED )
			||
			(isNonEmptyClientCertificate{Old}(state, message, len, direction) && \at(state->client_auth,New) == AUTH_PROVIDED )
		);

 predicate StateAfterClientKeyExchange{L}(STATE_MON *state) =
    \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
 	 	 \at(state->last_message,L) == C_KE										&&
		HaveSameStateValuesAndRole_E{L}(state, s)										&&
		MessageAddedToLog_E{L}(state, s, message, len) 							&&
		isClientKeyExchange{L}(message, len, handshake)							&&
		(
			(StateAfterClientCertificate(s) && (s->client_auth == AUTH_PROVIDED || s->client_auth == AUTH_NOT_PROVIDED))
			||
			(StateAfterServerHelloDone(s) && ( s->client_auth == NO_AUTH ))
		);

 predicate NextStateAfterClientKeyExchange{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len) =
 	 	 \at(state->last_message,New) == C_KE									&&
		HaveSameStateValuesAndRole{New,Old}(state)										&&
		MessageAddedToLog{New,Old}(state,message, len) 							&&
		isClientKeyExchange{New}(message, len, handshake)						&&
		(
			(StateAfterClientCertificate{Old}(state) && (\at(state->client_auth,Old) == AUTH_PROVIDED || \at(state->client_auth,Old) == AUTH_NOT_PROVIDED))
			||
			(StateAfterServerHelloDone{Old}(state) && ( \at(state->client_auth,Old) == NO_AUTH ))
		);

 predicate StateAfterClientCertificateVerify{L}(STATE_MON *state) =
    \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		\at(state->last_message,L) == C_CV										&&
		HaveSameStateValuesAndRole_E{L}(state, s)										&&
		MessageAddedToLog_E{L}(state, s, message, len) 							&&
		isClientCertificateVerify{L}(message, len, handshake)					&&
		StateAfterClientKeyExchange(s)											&&
		s->client_auth == AUTH_PROVIDED											;

 predicate NextStateAfterClientCertificateVerify{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len) =
		\at(state->last_message,New) == C_CV									&&
		HaveSameStateValuesAndRole{New,Old}(state)										&&
		MessageAddedToLog{New,Old}(state,message, len) 							&&
		isClientCertificateVerify{New}(message, len, handshake)					&&
		StateAfterClientKeyExchange{Old}(state)									&&
		\at(state->client_auth,Old) == AUTH_PROVIDED							;


 predicate StateAfterClientCCS{L}(STATE_MON *state) =
     \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		\at(state->last_message,L) == C_CCS										&&
		\at(state->resumption,L) == NO											&&
		HaveSameStateValuesButResumption_E{L}(state, s)							&&
		isClientChangeCipherSpecs{L}(state, message, len, direction, change_cipher_spec)	&&
		(
			(StateAfterClientCertificateVerify(s)  )
			||
			(StateAfterClientKeyExchange(s) && (s->client_auth == NO_AUTH || s->client_auth == AUTH_NOT_PROVIDED) )
			||
			(StateAfterClientCertificate(s) && (s->kem == DiffieHellman || s->kem == ECDH) )
		);

 predicate NextStateAfterClientCCS{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len, int direction) =
		\at(state->last_message,New) == C_CCS									&&
		\at(state->resumption,New) == NO										&&
		HaveSameStateValuesButResumption{New,Old}(state)						&&
		isClientChangeCipherSpecs{New}(state, message, len, direction, change_cipher_spec)&&
		(
			(StateAfterClientCertificateVerify{Old}(state)  )
			||
			(StateAfterClientKeyExchange{Old}(state) && (\at(state->client_auth,Old) == NO_AUTH || \at(state->client_auth,Old) == AUTH_NOT_PROVIDED ))
			||
			(StateAfterClientCertificate{Old}(state) && (\at(state->kem,Old) == DiffieHellman || \at(state->kem,Old) == ECDH) )
		);

 predicate StateAfterClientFin{L}(STATE_MON *state) =
    \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		\at(state->last_message,L) == C_Fin										&&
		HaveSameStateValuesAndRole_E{L}(state, s)										&&
		MessageAddedToLog_E{L}(state, s, message, len) 							&&
		isClientFin{L}(state, message, len, direction, handshake)				&&
//		s->resumption == NO							       	&&
		StateAfterClientCCS(s);

 predicate NextStateAfterClientFin{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len, int direction) =
		\at(state->last_message,New) == C_Fin									&&
		HaveSameStateValuesAndRole{New,Old}(state)										&&
		MessageAddedToLog{New,Old}(state,message, len) 							&&
		isClientFin{New}(state, message, len, direction, handshake)				&&
//		\at(state->resumption,Old) == NO										&&
		StateAfterClientCCS{Old}(state);

 predicate StateAfterServerNewSessionTicket{L}(STATE_MON *state) =
    \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		\at(state->last_message,L) == ServerNewSessionTicket					&&
		HaveSameStateValuesButResumptionAndTicket_E{L}(state, s)				&&
		MessageAddedToLog_E{L}(state, s, message, len) 							&&
		isServerNewSessionTicket{L}(message, len, handshake)					&&
		\at(state->ntick,L) == YES												&&
		(
			(StateAfterServerHello(s) && \at(state->resumption,L) == YES )
			||
			(StateAfterClientFin(s) && s->resumption == NO && s->resumption == \at(state->resumption,L) )
		);

 predicate NextStateAfterServerNewSessionTicket{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len) =
		\at(state->last_message,New) == ServerNewSessionTicket					&&
		HaveSameStateValuesButResumptionAndTicket{New,Old}(state)				&&
		MessageAddedToLog{New,Old}(state,message, len) 							&&
		isServerNewSessionTicket{New}(message, len, handshake)					&&
		\at(state->ntick,New) == YES											&&
		(
			(StateAfterServerHello{Old}(state) && \at(state->resumption,New) == YES )
			||
			(StateAfterClientFin{Old}(state) && \at(state->resumption,Old) == NO && \at(state->resumption,Old) == \at(state->resumption,New) )
		);

 predicate StateAfterServerCCS{L}(STATE_MON *state) =
    \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		\at(state->last_message,L) == S_CCS										&&
		HaveSameStateValuesButResumptionAndTicket_E{L}(state, s)				&&
		isServerChangeCipherSpecs{L}(state, message, len, direction, change_cipher_spec)	&&
		(
			(StateAfterServerHello(s) && \at(state->resumption,L) == YES && \at(state->ntick,L) == NO )
			||
			(StateAfterClientFin(s) && s->resumption == NO && s->resumption == \at(state->resumption,L)  && s->ntick == UNDEFINED )
			||
			(StateAfterServerNewSessionTicket(s) && \at(state->resumption,L) == s->resumption  && \at(state->ntick,L) == s->ntick )
		);

 predicate NextStateAfterServerCCS{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len, int direction) =
		\at(state->last_message,New) == S_CCS									&&
		HaveSameStateValuesButResumptionAndTicket{New,Old}(state)				&&
		isServerChangeCipherSpecs{New}(state, message, len, direction, change_cipher_spec)&&
		(
			(StateAfterServerHello{Old}(state) && \at(state->resumption,New) == YES && \at(state->ntick,New) == NO )
			||
			(StateAfterClientFin{Old}(state) && \at(state->resumption,Old) == NO && \at(state->resumption,Old) == \at(state->resumption,New)  && \at(state->ntick,New) == NO )
			||
			(StateAfterServerNewSessionTicket{Old}(state) && \at(state->resumption,New) == \at(state->resumption,Old) && \at(state->ntick,New) == \at(state->ntick,Old))
		);

 predicate StateAfterServerFin{L}(STATE_MON *state) =
    \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		\at(state->last_message,L) == S_Fin										&&
		HaveSameStateValuesAndRole_E{L}(state, s)										&&
		MessageAddedToLog_E{L}(state, s, message, len) 							&&
		isServerFin{L}(state, message, len, direction, handshake)				&&
		StateAfterServerCCS(s);

 predicate NextStateAfterServerFin{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len, int direction) =
		\at(state->last_message,New) == S_Fin									&&
		HaveSameStateValuesAndRole{New,Old}(state)										&&
		MessageAddedToLog{New,Old}(state,message, len) 							&&
		isServerFin{New}(state, message, len, direction, handshake)				&&
		StateAfterServerCCS{Old}(state);

 predicate StateAfterClientCCSLastMsg{L}(STATE_MON *state) =
     \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		\at(state->last_message,L) == C_CCS										&&
		HaveSameStateValuesAndRole_E{L}(state, s)										&&
		isClientChangeCipherSpecs{L}(state, message, len, direction, change_cipher_spec)	&&
		\at(state->resumption,L) == YES											&&
		StateAfterServerFin(s) ;

 predicate NextStateAfterClientCCSLastMsg{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len, int direction) =
		\at(state->last_message,New) == C_CCS									&&
		HaveSameStateValuesAndRole{New,Old}(state)										&&
		isClientChangeCipherSpecs{New}(state, message, len, direction, change_cipher_spec)&&
		\at(state->resumption,New) == YES										&&
		StateAfterServerFin{Old}(state) ;

 predicate StateAfterClientFinLastMsg{L}(STATE_MON *state) =
    \exists STATE_MON *s, unsigned char *message, unsigned int len, int direction;
		\at(state->last_message,L) == C_Fin										&&
		HaveSameStateValuesAndRole_E{L}(state, s)										&&
		MessageAddedToLog_E{L}(state, s, message, len) 							&&
		isClientFin{L}(state, message, len, direction, handshake)				&&
//		s->resumption == YES               &&  
		StateAfterClientCCSLastMsg(s)	;


 predicate NextStateAfterClientFinLastMsg{New,Old}(STATE_MON *state, unsigned char *message, unsigned int len, int direction) =
		\at(state->last_message,New) == C_Fin									&&
		HaveSameStateValuesAndRole{New,Old}(state)										&&
		MessageAddedToLog{New,Old}(state,message, len) 							&&
		isClientFin{New}(state, message, len, direction, handshake)				&&
//		\at(state->resumption,Old) == YES &&
		StateAfterClientCCSLastMsg{Old}(state) ;

 predicate isValidState{L}(STATE_MON *state) =
 	 StateAfterInitialState{L}(state) 				||
 	 StateAfterClientHello{L}(state) 				||
 	 StateAfterServerHello{L}(state) 				||
 	 StateAfterServerCertificate{L}(state) 			||
 	 StateAfterServerKeyExchange{L}(state) 			||
 	 StateAfterServerCertificateRequest{L}(state) 	||
 	 StateAfterServerHelloDone{L}(state) 			||
 	 StateAfterClientCertificate{L}(state) 			||
 	 StateAfterClientKeyExchange{L}(state)			||
 	 StateAfterClientCertificateVerify{L}(state) 	||
 	 StateAfterServerNewSessionTicket{L}(state) 	||
 	 StateAfterServerCCS{L}(state) 					||
 	 StateAfterServerFin{L}(state) 					||
 	 StateAfterClientCCS{L}(state) 					||
 	 StateAfterClientFin{L}(state)					||
 	 StateAfterClientCCSLastMsg{L}(state) 			||
 	 StateAfterClientFinLastMsg{L}(state)			;

//===================================================================
// 				Axioms between the instanciated and
// 				non instanciated state monitor predicates
//===================================================================

axiom state_after_client_hello{L1,L2} :
	\forall STATE_MON *state, unsigned char *msg, unsigned int len;
		NextStateAfterClientHello{L1,L2}(state, msg, len) ==> StateAfterClientHello{L1}(state);

axiom state_after_server_hello{L1,L2} :
	\forall STATE_MON *state, unsigned char *msg, unsigned int len;
		NextStateAfterServerHello{L1,L2}(state, msg, len) ==> StateAfterServerHello{L1}(state);

axiom state_after_server_certificate{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len, int direction;
		NextStateAfterServerCertificate{L1,L2}(state, msg, len, direction) ==> StateAfterServerCertificate{L1}(state);

axiom state_after_server_key_exchange{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len;
		NextStateAfterServerKeyExchange{L1,L2}(state, msg, len) ==> StateAfterServerKeyExchange{L1}(state);

axiom state_after_server_certificate_request{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len;
		NextStateAfterServerCertificateRequest{L1,L2}(state, msg, len) ==> StateAfterServerCertificateRequest{L1}(state);

axiom state_after_server_hello_done{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len;
		NextStateAfterServerHelloDone{L1,L2}(state, msg, len) ==> StateAfterServerHelloDone{L1}(state);

axiom state_after_client_certificate{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len, int direction;
		NextStateAfterClientCertificate{L1,L2}(state, msg, len, direction) ==> StateAfterClientCertificate{L1}(state);

axiom state_after_client_key_exchange{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len;
		NextStateAfterClientKeyExchange{L1,L2}(state, msg, len) ==> StateAfterClientKeyExchange{L1}(state);

axiom state_after_client_certificate_verify{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len;
		NextStateAfterClientCertificateVerify{L1,L2}(state, msg, len) ==> StateAfterClientCertificateVerify{L1}(state);

axiom state_after_server_new_session_ticket{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len;
		NextStateAfterServerNewSessionTicket{L1,L2}(state, msg, len) ==> StateAfterServerNewSessionTicket{L1}(state);

axiom state_after_server_change_cipher_spec{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len, int direction;
		NextStateAfterServerCCS{L1,L2}(state, msg, len, direction) ==> StateAfterServerCCS{L1}(state);

axiom state_after_server_finished{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len, int direction;
		NextStateAfterServerFin{L1,L2}(state, msg, len, direction) ==> StateAfterServerFin{L1}(state);

axiom state_after_client_change_cipher_spec{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len, int direction;
		NextStateAfterClientCCS{L1,L2}(state, msg, len, direction) ==> StateAfterClientCCS{L1}(state);

axiom state_after_client_finished{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len, int direction;
		NextStateAfterClientFin{L1,L2}(state, msg, len, direction) ==> StateAfterClientFin{L1}(state);

axiom state_after_client_change_cipher_spec_last_msg{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len, int direction;
		NextStateAfterClientCCSLastMsg{L1,L2}(state, msg, len, direction) ==> StateAfterClientCCSLastMsg{L1}(state);

axiom state_after_client_finished_last_msg{L1,L2}:
	\forall STATE_MON *state, unsigned char *msg, unsigned int len, int direction;
		NextStateAfterClientFinLastMsg{L1,L2}(state, msg, len, direction) ==> StateAfterClientFinLastMsg{L1}(state);


 predicate HaveSameStateValues{L1,L2}(STATE_MON *s) =
	\at(s->version,L1) 			== \at(s->version,L2) 				&&
	\at(s->kem,L1) 				== \at(s->kem,L2) 					&&
	\at(s->client_auth,L1) 		== \at(s->client_auth,L2) 			&&
	\at(s->resumption,L1) 		== \at(s->resumption,L2) 			&&
	\at(s->renegotiation,L1) 	== \at(s->renegotiation,L2) 		&&
	\at(s->ntick, L1) 			== \at(s->ntick, L2)				;


}

*/

#endif /* STATE_MACHINE_H_ */
