/*@
 axiomatic PARSING {

//===================================================================
// 				Predicates on who sent the message
//===================================================================

 predicate sentByClient(STATE_MON *state, int direction) =
 	 (state->role == SERVER && direction == RECEIVE)||
 	 (state->role == CLIENT && direction == SEND);

 predicate sentByServer(STATE_MON *state, int direction) =
 	 (state->role == CLIENT && direction == RECEIVE)||
 	 (state->role == SERVER && direction == SEND);

//===================================================================
// 				Predicates on the nature of the message
//===================================================================

// predicate isMessageLength(unsigned char *msg, unsigned int len) =
// 	 len >= 4 && (len - 4) == (65536 * msg[1] + 256 * msg[2] + msg[3]);
predicate isMessageLength(unsigned char *msg, unsigned int len) =
 	 len >= 4 && (len-4) == (65536 * msg[1] + 256 * msg[2] + msg[3]);

 predicate isClientHello(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == C_Hello &&
 	 isMessageLength(msg, len);
 predicate isServerHello(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == S_Hello &&
 	 isMessageLength(msg, len);
 predicate isServerCertificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == handshake && msg[0] == certificate && sentByServer(state, direction)  &&
 	 isMessageLength(msg, len);
 predicate isServerKeyExchange(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == S_KE &&
 	 isMessageLength(msg, len);
 predicate isServerCertificateRequest(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == S_CReq &&
 	 isMessageLength(msg, len);
 predicate isServerHelloDone(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == S_HD &&
 	 isMessageLength(msg, len);
 predicate isClientCertificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == handshake && msg[0] == certificate && sentByClient(state, direction)  &&
 	 isMessageLength(msg, len);
 predicate isClientKeyExchange(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == C_KE &&
 	 isMessageLength(msg, len);
 predicate isClientCertificateVerify(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == C_CV &&
 	 isMessageLength(msg, len);
 predicate isServerChangeCipherSpecs(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == change_cipher_spec && sentByServer(state, direction) &&
 	 len == 1 && msg[0] == 0x01;
 predicate isServerFin(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == handshake && msg[0] == finished && sentByServer(state, direction)  &&
 	 isMessageLength(msg, len);
 predicate isClientChangeCipherSpecs(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == change_cipher_spec && sentByClient(state, direction) &&
 	 len == 1 && msg[0] == 0x01;
 predicate isClientFin(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 contentType == handshake && msg[0] == finished && sentByClient(state, direction) &&
 	 isMessageLength(msg, len);
 predicate isServerNewSessionTicket(unsigned char *msg, unsigned int len, unsigned char contentType) =
 	 contentType == handshake && msg[0] == ServerNewSessionTicket &&
 	 isMessageLength(msg, len);

 predicate isUnknownMessage(STATE_MON *state, unsigned char *msg, unsigned int len, int direction, unsigned char contentType) =
 	 !isClientHello(msg, len, contentType) 									&&
 	 !isServerHello(msg, len, contentType) 									&&
	 !isServerCertificate(state, msg, len, direction, contentType) 			&&
	 !isServerKeyExchange(msg, len, contentType) 							&&
	 !isServerCertificateRequest(msg, len, contentType) 					&&
	 !isServerHelloDone(msg, len, contentType) 								&&
	 !isClientCertificate(state, msg, len, direction, contentType) 			&&
	 !isClientKeyExchange(msg, len, contentType) 							&&
	 !isClientCertificateVerify(msg, len, contentType) 						&&
	 !isServerChangeCipherSpecs(state, msg, len, direction, contentType) 	&&
	 !isServerFin(state, msg, len, direction, contentType) 					&&
	 !isClientChangeCipherSpecs(state, msg, len, direction, contentType) 	&&
	 !isServerNewSessionTicket(msg, len, contentType) 						&&
	 !isClientFin(state, msg, len, direction, contentType)					;

 predicate isEmptyClientCertificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction) =
 	 isClientCertificate(state, msg, len, direction, handshake) 		&&
 	 len == 4;

 predicate isNonEmptyClientCertificate(STATE_MON *state, unsigned char *msg, unsigned int len, int direction) =
 	 isClientCertificate(state, msg, len, direction, handshake)			&&
 	 len > 4;

//===================================================================
// 				Predicates on the protocol versions
//===================================================================

 predicate isSSL3(unsigned char *msg, unsigned int len)=
  	 len >= 6 &&
 	 msg[4] == 3 && msg[5] == 0;
 predicate isTLS1_0(unsigned char *msg, unsigned int len)=
   	 len >= 6 &&
 	 msg[4] == 3 && msg[5] == 1;
 predicate isTLS1_1(unsigned char *msg, unsigned int len)=
   	 len >= 6 &&
 	 msg[4] == 3 && msg[5] == 2;
 predicate isTLS1_2(unsigned char *msg, unsigned int len)=
   	 len >= 6 &&
 	 msg[4] == 3 && msg[5] == 3;

 predicate isUnknownPV(unsigned char *msg, unsigned int len) =
 	 !isSSL3(msg, len) 		&&
 	 !isTLS1_0(msg, len) 	&&
 	 !isTLS1_1(msg, len) 	&&
 	 !isTLS1_2(msg, len);

//===================================================================
// 				Predicates on the key exchange methods
//===================================================================

 predicate isDH(unsigned char *msg, unsigned int len) =
   	 len >= 39 + msg[38]+2 &&
   	 \let offset = 39 + msg[38];
 	 (
 	 	 ( msg[offset] == 0x00 &&
 	 	 	 (
				msg[offset+1] ==  0x0C ||
				msg[offset+1] ==  0x0D ||
				msg[offset+1] ==  0x0F ||
				msg[offset+1] ==  0x10 ||
				msg[offset+1] ==  0x30 ||
				msg[offset+1] ==  0x31 ||
				msg[offset+1] ==  0x36 ||
				msg[offset+1] ==  0x37 ||
				msg[offset+1] ==  0x3E ||
				msg[offset+1] ==  0x3F ||
				msg[offset+1] ==  0x42 ||
				msg[offset+1] ==  0x43 ||
				msg[offset+1] ==  0x68 ||
				msg[offset+1] ==  0x69 ||
				msg[offset+1] ==  0x85 ||
				msg[offset+1] ==  0x86 ||
				msg[offset+1] ==  0x97 ||
				msg[offset+1] ==  0x98 ||
				msg[offset+1] ==  0xA0 ||
				msg[offset+1] ==  0xA1 ||
				msg[offset+1] ==  0xA4 ||
				msg[offset+1] ==  0xA5 ||
				msg[offset+1] ==  0xBB ||
				msg[offset+1] ==  0xBC ||
				msg[offset+1] ==  0xC1 ||
				msg[offset+1] ==  0xC2
 	 	 	 )
 	 	 )
 	 	 ||
 	 	 ( msg[offset] == 0xC0 &&
 	 	 	 (
				msg[offset+1] ==  0x3E ||
				msg[offset+1] ==  0x3F ||
				msg[offset+1] ==  0x40 ||
				msg[offset+1] ==  0x41 ||
				msg[offset+1] ==  0x54 ||
				msg[offset+1] ==  0x55 ||
				msg[offset+1] ==  0x58 ||
				msg[offset+1] ==  0x59 ||
				msg[offset+1] ==  0x7E ||
				msg[offset+1] ==  0x7F ||
				msg[offset+1] ==  0x82 ||
				msg[offset+1] ==  0x83
 	 	 	 )
 	 	 )
 	 );
 predicate isDHE(unsigned char *msg, unsigned int len) =
   	 len >= 39 + msg[38]+2 &&
   	 \let offset = 39 + msg[38];
 	 (
 	 	 ( msg[offset] == 0x00 &&
 	 	 	 (
				msg[offset+1] ==  0x12 ||
				msg[offset+1] ==  0x13 ||
				msg[offset+1] ==  0x15 ||
				msg[offset+1] ==  0x16 ||
				msg[offset+1] ==  0x32 ||
				msg[offset+1] ==  0x33 ||
				msg[offset+1] ==  0x38 ||
				msg[offset+1] ==  0x39 ||
				msg[offset+1] ==  0x40 ||
				msg[offset+1] ==  0x44 ||
				msg[offset+1] ==  0x45 ||
				msg[offset+1] ==  0x67 ||
				msg[offset+1] ==  0x6A ||
				msg[offset+1] ==  0x6B ||
				msg[offset+1] ==  0x87 ||
				msg[offset+1] ==  0x88 ||
				msg[offset+1] ==  0x99 ||
				msg[offset+1] ==  0x9A ||
				msg[offset+1] ==  0x9E ||
				msg[offset+1] ==  0x9F ||
				msg[offset+1] ==  0xA2 ||
				msg[offset+1] ==  0xA3 ||
				msg[offset+1] ==  0xBD ||
				msg[offset+1] ==  0xBE ||
				msg[offset+1] ==  0xC3 ||
				msg[offset+1] ==  0xC4
 	 	 	 )
 	 	 )
 	 	 ||
 	 	 ( msg[offset] == 0xC0 &&
 	 	 	 (
				msg[offset+1] ==  0x42 ||
				msg[offset+1] ==  0x43 ||
				msg[offset+1] ==  0x44 ||
				msg[offset+1] ==  0x45 ||
				msg[offset+1] ==  0x52 ||
				msg[offset+1] ==  0x53 ||
				msg[offset+1] ==  0x56 ||
				msg[offset+1] ==  0x57 ||
				msg[offset+1] ==  0x7C ||
				msg[offset+1] ==  0x7D ||
				msg[offset+1] ==  0x80 ||
				msg[offset+1] ==  0x81 ||
				msg[offset+1] ==  0x9E ||
				msg[offset+1] ==  0x9F ||
				msg[offset+1] ==  0xA2 ||
				msg[offset+1] ==  0xA3
 	 	 	 )
 	 	 )
 	 );
 predicate isDHEExport(unsigned char *msg, unsigned int len) =
   	 len >= 39 + msg[38]+2 &&
   	 \let offset = 39 + msg[38];
 	 (
 	 	 ( msg[offset] == 0x00 &&
 	 	 	 (
				msg[offset+1] ==  0x11 ||
				msg[offset+1] ==  0x14
 	 	 	 )
 	 	 )
 	 );
 predicate isDHanon(unsigned char *msg, unsigned int len) =
   	 len >= 39 + msg[38]+2 &&
   	 \let offset = 39 + msg[38];
 	 (
 	 	 ( msg[offset] == 0x00 &&
 	 	 	 (
				msg[offset+1] == 0x18 ||
				msg[offset+1] == 0x1A ||
				msg[offset+1] == 0x1B ||
				msg[offset+1] == 0x34 ||
				msg[offset+1] == 0x3A ||
				msg[offset+1] == 0x46 ||
				msg[offset+1] == 0x6C ||
				msg[offset+1] == 0x6D ||
				msg[offset+1] == 0x89 ||
				msg[offset+1] == 0x9B ||
				msg[offset+1] == 0xA6 ||
				msg[offset+1] == 0xA7 ||
				msg[offset+1] == 0xBF ||
				msg[offset+1] == 0xC5
 	 	 	 )
 	 	 )
 	 	 ||
 	 	 ( msg[offset] == 0xC0 &&
 	 	 	 (
				msg[offset+1] == 0x46 ||
				msg[offset+1] == 0x47 ||
				msg[offset+1] == 0x5A ||
				msg[offset+1] == 0x5B ||
				msg[offset+1] == 0x84 ||
				msg[offset+1] == 0x85
 	 	 	 )
 	 	 )
 	 );
 predicate isECDH(unsigned char *msg, unsigned int len)=
   	 len >= 39 + msg[38] + 2 &&
   	 \let offset = 39 + msg[38];
 	 (
 	 	 ( msg[offset] == 0xC0 &&
 	 	 	 (
				msg[offset+1] ==  0x01 ||
				msg[offset+1] ==  0x02 ||
				msg[offset+1] ==  0x03 ||
				msg[offset+1] ==  0x04 ||
				msg[offset+1] ==  0x05 ||
				msg[offset+1] ==  0x0B ||
				msg[offset+1] ==  0x0C ||
				msg[offset+1] ==  0x0D ||
				msg[offset+1] ==  0x0E ||
				msg[offset+1] ==  0x0F ||
				msg[offset+1] ==  0x25 ||
				msg[offset+1] ==  0x26 ||
				msg[offset+1] ==  0x29 ||
				msg[offset+1] ==  0x2A ||
				msg[offset+1] ==  0x2D ||
				msg[offset+1] ==  0x2E ||
				msg[offset+1] ==  0x31 ||
				msg[offset+1] ==  0x32 ||
				msg[offset+1] ==  0x4A ||
				msg[offset+1] ==  0x4B ||
				msg[offset+1] ==  0x4E ||
				msg[offset+1] ==  0x4F ||
				msg[offset+1] ==  0x5E ||
				msg[offset+1] ==  0x5F ||
				msg[offset+1] ==  0x62 ||
				msg[offset+1] ==  0x63 ||
				msg[offset+1] ==  0x74 ||
				msg[offset+1] ==  0x75 ||
				msg[offset+1] ==  0x78 ||
				msg[offset+1] ==  0x79 ||
				msg[offset+1] ==  0x88 ||
				msg[offset+1] ==  0x89 ||
				msg[offset+1] ==  0x8C ||
				msg[offset+1] ==  0x8D
 	 	 	 )
 	 	 )
 	 );
 predicate isECDHE(unsigned char *msg, unsigned int len)=
   	 len >= 39 + msg[38]+2 &&
   	 \let offset = 39 + msg[38];
 	 (
 	 	 ( msg[offset] == 0xC0 &&
 	 	 	 (
				msg[offset+1] ==  0x06 ||
				msg[offset+1] ==  0x07 ||
				msg[offset+1] ==  0x08 ||
				msg[offset+1] ==  0x09 ||
				msg[offset+1] ==  0x0A ||
				msg[offset+1] ==  0x10 ||
				msg[offset+1] ==  0x11 ||
				msg[offset+1] ==  0x12 ||
				msg[offset+1] ==  0x13 ||
				msg[offset+1] ==  0x14 ||
				msg[offset+1] ==  0x23 ||
				msg[offset+1] ==  0x24 ||
				msg[offset+1] ==  0x27 ||
				msg[offset+1] ==  0x28 ||
				msg[offset+1] ==  0x2B ||
				msg[offset+1] ==  0x2C ||
				msg[offset+1] ==  0x2F ||
				msg[offset+1] ==  0x30 ||
				msg[offset+1] ==  0x48 ||
				msg[offset+1] ==  0x49 ||
				msg[offset+1] ==  0x4C ||
				msg[offset+1] ==  0x4D ||
				msg[offset+1] ==  0x5C ||
				msg[offset+1] ==  0x5D ||
				msg[offset+1] ==  0x60 ||
				msg[offset+1] ==  0x61 ||
				msg[offset+1] ==  0x72 ||
				msg[offset+1] ==  0x73 ||
				msg[offset+1] ==  0x76 ||
				msg[offset+1] ==  0x77 ||
				msg[offset+1] ==  0x8A ||
				msg[offset+1] ==  0x8B ||
				msg[offset+1] ==  0xAc ||
				msg[offset+1] ==  0xAD ||
				msg[offset+1] ==  0xAE ||
				msg[offset+1] ==  0xAF
 	 	 	 )
 	 	 )
 	 );
 predicate isECDHanon(unsigned char *msg, unsigned int len) =
   	 len >= 39 + msg[38]+2 &&
   	 \let offset = 39 + msg[38];
 	 (
 	 	 ( msg[offset] == 0xC0 &&
 	 	 	 (
				msg[offset+1] ==  0x15 ||
				msg[offset+1] ==  0x16 ||
				msg[offset+1] ==  0x17 ||
				msg[offset+1] ==  0x18 ||
				msg[offset+1] ==  0x19
 	 	 	 )
 	 	 )
 	 );
 predicate isRSA(unsigned char *msg, unsigned int len)=
   	 len >= 39 + msg[38]+2 &&
   	 \let offset = 39 + msg[38];
 	 (
 	 	 ( msg[offset] == 0x00 &&
 	 	 	 (
 	 	 	 	msg[offset+1] == 0x01 ||
				msg[offset+1] == 0x02 ||
				msg[offset+1] == 0x3B ||
				msg[offset+1] == 0x04 ||
				msg[offset+1] == 0x05 ||
				msg[offset+1] == 0x07 ||
				msg[offset+1] == 0x09 ||
				msg[offset+1] == 0x0A ||
				msg[offset+1] == 0x2F ||
				msg[offset+1] == 0x35 ||
				msg[offset+1] == 0x3C ||
				msg[offset+1] == 0x3D ||
				msg[offset+1] == 0x84 ||
				msg[offset+1] == 0x96 ||
				msg[offset+1] == 0x9C ||
				msg[offset+1] == 0x9D ||
				msg[offset+1] == 0xBA ||
				msg[offset+1] == 0xC0
 	 	 	 )
 	 	 )
 	 	 ||
 	 	 ( msg[offset] == 0xC0 &&
 	 	 	 (
 	 	 	 	msg[offset+1] ==  0x3C ||
				msg[offset+1] ==  0x3D ||
				msg[offset+1] ==  0x50 ||
				msg[offset+1] ==  0x51 ||
				msg[offset+1] ==  0x7A ||
				msg[offset+1] ==  0x7B ||
				msg[offset+1] ==  0x9C ||
				msg[offset+1] ==  0x9D ||
				msg[offset+1] ==  0xA0 ||
				msg[offset+1] ==  0xA1
 	 	 	 )
 	 	 )
 	 );
 predicate isRSAExport(unsigned char *msg, unsigned int len)=
   	 len >= 39 + msg[38]+2 &&
   	 \let offset = 39 + msg[38];
 	 (
 	 	 ( msg[offset] == 0x00 &&
 	 	 	 (
 	 	 	 	msg[offset+1] == 0x03 ||
				msg[offset+1] == 0x06 ||
				msg[offset+1] == 0x08
 	 	 	 )
 	 	 )
 	 );


 predicate isUnknownCS(unsigned char *msg, unsigned int len) =
 	 !isDH(msg, len) 		&&
 	 !isDHE(msg, len)		&&
 	 !isDHEExport(msg, len)	&&
 	 !isDHanon(msg, len) 	&&
 	 !isECDH(msg, len) 		&&
 	 !isECDHE(msg, len) 	&&
 	 !isECDHanon(msg, len) 	&&
 	 !isRSA(msg, len) 		&&
 	 !isRSAExport(msg, len) ;

 }
 */


/*@
 assigns \nothing;

 ensures (isClientHello(msg, len, contentType) 								<==> \result == C_Hello) 				&&
 	 	 (isServerHello(msg, len, contentType) 								<==> \result == S_Hello )				&&
 	 	 (isServerCertificate(state, msg, len, direction, contentType) 		<==> \result == S_Cert) 				&&
 	 	 (isServerKeyExchange(msg, len, contentType) 						<==> \result == S_KE) 					&&
 	 	 (isServerCertificateRequest(msg, len, contentType) 				<==> \result == S_CReq) 				&&
 	 	 (isServerHelloDone(msg, len, contentType) 							<==> \result == S_HD )					&&
 	 	 (isClientCertificate(state, msg, len, direction, contentType) 		<==> \result == C_Cert) 				&&
 	 	 (isClientKeyExchange(msg, len, contentType) 						<==> \result == C_KE) 					&&
 	 	 (isClientCertificateVerify(msg, len, contentType) 					<==> \result == C_CV )					&&
 	 	 (isServerChangeCipherSpecs(state, msg, len, direction, contentType)<==> \result ==	S_CCS) 					&&
 	 	 (isServerFin(state, msg, len, direction, contentType) 				<==> \result == S_Fin) 					&&
 	 	 (isClientChangeCipherSpecs(state, msg, len, direction, contentType)<==> \result == C_CCS )					&&
 	 	 (isClientFin(state, msg, len, direction, contentType) 				<==> \result == C_Fin) 					&&
 	 	 (isServerNewSessionTicket(msg, len, contentType) 					<==> \result == ServerNewSessionTicket) &&
 	 	 (isUnknownMessage(state, msg, len, direction, contentType)			<==> \result == UNDEFINED_TYPE )		;
   */
Msg_type getMessageType(STATE_MON *state, unsigned char *msg, unsigned int len,
		int direction, unsigned char contentType);

/*@
 requires isMessageLength(msg, len);

 assigns \nothing;

 ensures
 	 (\result == SSL3 <==> isSSL3(msg, len)) &&
 	 (\result == TLS1 <==> isTLS1_0(msg, len)) &&
 	 (\result == TLS1_1 <==> isTLS1_1(msg, len)) &&
 	 (\result == TLS1_2 <==> isTLS1_2(msg, len)) &&
 	 //(\result == TICKETS <==> isTickets(msg, len)) &&
 	 //(\result == ALPN <==> isALPN(msg, len)) &&
 	 //(\result == DTLS <==> isDTLS(msg, len)) &&
 	 (\result == UNDEFINED_PV <==> isUnknownPV(msg, len));
 */
PV getPV(unsigned char *msg, unsigned int len);

/*@
 requires isMessageLength(msg, len);

 assigns \nothing;

  ensures
 	 (\result == DiffieHellman 	<==> isDH(msg, len))	  		&&
 	 (\result == DHE 			<==> isDHE(msg, len)) 			&&
 	 (\result == DHEExport		<==> isDHEExport(msg, len))		&&
 	 (\result == DHanon 		<==> isDHanon(msg, len))		&&
 	 (\result == ECDH 			<==> isECDH(msg, len))  		&&
 	 (\result == ECDHE 			<==> isECDHE(msg, len)) 		&&
 	 (\result == ECDHanon 		<==> isECDHanon(msg, len)) 		&&
 	 (\result == CS_RSA 		<==> isRSA(msg, len)) 			&&
 	 (\result == RSAExport 		<==> isRSAExport(msg, len))  	&&
 	 (\result == UNDEFINED_KEM 	<==> isUnknownCS(msg, len)) 	 ;
 */
KEM getKEM(unsigned char *msg, unsigned int len);

/*@
  requires isClientCertificate(state, msg, len, direction, handshake);

  assigns \nothing;
  behavior other_certificate:
  	  assumes (msg[1] != 0 || msg[2] != 0 || msg[3] != 0);
  	  ensures \result == AUTH_PROVIDED <==> isNonEmptyClientCertificate{Pre}(state, msg, len, direction);
  behavior empty_certificate :
  	  assumes msg[1] == 0 && msg[2] == 0 && msg[3] == 0;
  	  ensures \result == AUTH_NOT_PROVIDED <==> isEmptyClientCertificate{Pre}(state, msg, len, direction);
  complete behaviors other_certificate, empty_certificate;
  disjoint behaviors other_certificate, empty_certificate;
 */
Auth checkClientCertificate(STATE_MON *state, unsigned char *msg,
		unsigned int len, int direction);
